package au.com.dius.pact.core.matchers

import au.com.dius.pact.core.model.OptionalBody
import au.com.dius.pact.core.model.matchingrules.MatchingRulesImpl
import au.com.dius.pact.core.model.matchingrules.TypeMatcher
import spock.lang.Specification

@SuppressWarnings(['PrivateFieldCouldBeFinal', 'LineLength'])
class FormPostBodyMatcherSpec extends Specification {

  private FormPostBodyMatcher matcher
  private MatchingRulesImpl matchers

  def setup() {
    matcher = new FormPostBodyMatcher()
    matchers = new MatchingRulesImpl()
  }

  def 'returns no mismatches - when the expected body is missing'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches.empty

    where:
    actualBody = OptionalBody.empty()
    expectedBody = OptionalBody.missing()
  }

  def 'returns no mismatches - when the expected body and actual bodies are empty'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches.empty

    where:
    actualBody = OptionalBody.empty()
    expectedBody = OptionalBody.empty()
  }

  def 'returns no mismatches - when the expected body and actual bodies are equal'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches.empty

    where:
    actualBody = OptionalBody.body('a=b&c=d'.bytes)
    expectedBody = OptionalBody.body('a=b&c=d'.bytes)
  }

  def 'returns no mismatches - when the actual body has extra keys and we allow unexpected keys'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches.empty

    where:
    actualBody = OptionalBody.body('a=b&c=d'.bytes)
    expectedBody = OptionalBody.body('a=b'.bytes)
  }

  def 'returns no mismatches - when the keys are in different order'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches.empty

    where:
    actualBody = OptionalBody.body('a=b&c=d'.bytes)
    expectedBody = OptionalBody.body('c=d&a=b'.bytes)
  }

  def 'returns mismatches - when the expected body contains keys that are not in the actual body'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches*.mismatch ==
      ['Expected form post parameter \'c\' but was missing']

    where:
    actualBody = OptionalBody.body('a=b'.bytes)
    expectedBody = OptionalBody.body('a=b&c=d'.bytes)
  }

  @SuppressWarnings('LineLength')
  def 'returns mismatches - when the actual body contains keys that are not in the expected body and we do not allow extra keys'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, false, matchers).mismatches*.mismatch ==
      ['Received unexpected form post parameter \'a\'=[\'b\']']

    where:
    actualBody = OptionalBody.body('a=b&c=d'.bytes)
    expectedBody = OptionalBody.body('c=d'.bytes)
  }

  def 'returns mismatches - when the expected body is present but there is no actual body'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches*.mismatch ==
      ['Expected a form post body but was missing']

    where:
    actualBody = OptionalBody.missing()
    expectedBody = OptionalBody.body('a=a'.bytes)
  }

  def 'returns mismatches - if the same key is repeated with values in different order'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches*.mismatch ==
      [
        'Expected form post parameter \'a\'[0] with value \'1\' but was \'2\'',
        'Expected form post parameter \'a\'[1] with value \'2\' but was \'1\''
      ]

    where:
    actualBody = OptionalBody.body('a=2&a=1&b=3'.bytes)
    expectedBody = OptionalBody.body('a=1&a=2&b=3'.bytes)
  }

  def 'returns mismatches - if the same key is repeated with values missing'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches*.mismatch ==
      [
        'Expected form post parameter \'a\'=\'3\' but was missing'
      ]

    where:
    actualBody = OptionalBody.body('a=1&a=2'.bytes)
    expectedBody = OptionalBody.body('a=1&a=2&a=3'.bytes)
  }

  def 'returns mismatches - when the actual body contains values that are not the same as the expected body'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches*.mismatch ==
      ['Expected form post parameter \'c\'[0] with value \'d\' but was \'1\'']

    where:
    actualBody = OptionalBody.body('a=b&c=1'.bytes)
    expectedBody = OptionalBody.body('c=d&a=b'.bytes)
  }

  def 'handles delimiters in the values'() {
    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches*.mismatch ==
      ['Expected form post parameter \'c\'[0] with value \'1\' but was \'1=2\'']

    where:
    actualBody = OptionalBody.body('a=b&c=1=2'.bytes)
    expectedBody = OptionalBody.body('c=1&a=b'.bytes)
  }

  def 'delegates to any defined matcher'() {
    given:
    matchers.addCategory('body').addRule('$.c', TypeMatcher.INSTANCE)

    expect:
    matcher.matchBody(expectedBody, actualBody, true, matchers).mismatches.empty

    where:
    actualBody = OptionalBody.body('a=b&c=2'.bytes)
    expectedBody = OptionalBody.body('c=1&a=b'.bytes)
  }
}
